"""
Interface to Mongo for using Mongo as a buffer for chunked data transfer.

We use Mongo as a buffer because we cannot guarantee that all of the requests
for individual chunks will hit the same worker. While we could use streamed responses for the download,
they're not available for upload, so for simplicity we use the same approach in both directions.

The intended usage pattern is that a single message is stored as a single "chunk set".
The chunk set is first "filled" (either by the client or the server), then "emptied" to obtain the data
(again, by either the client or the server).

Note that while this can be used for pagination, it is not in itself pagination.

We plan to gather most of this material into ew_common after the chunked transfer interface has been rolled out
to its three main users (ew_geostore, ew_specklia, ew_online_processing_service) and proven effective for each.
At that point, this entire module will move into ew_common. Note that the chunked transfer interface will always
require MongoDB or a similar provision to work correctly.

IMPORTANT: THE VERSION HERE IN THE SPECKLIA PACKAGE MUST NOT BE MADE DEPENDENT UPON EW_COMMON SINCE EW_COMMON
IS PRIVATE BUT THIS PACKAGE IS PUBLIC!
"""

import struct
import time
from enum import Enum
from io import BytesIO
from logging import Logger
from typing import List, Tuple, Union

import requests
from geopandas import GeoDataFrame
from geopandas import read_feather as read_geofeather
from pandas import DataFrame, read_feather

log = Logger(__name__)

CHUNK_DB_NAME = "data_transfer_chunks"
CHUNK_METADATA_COLLECTION_NAME = "chunk_metadata"
MAX_CHUNK_AGE_SECONDS = 3600
MAX_CHUNK_SIZE_BYTES = 5 * 1024**2  # must be small enough to fit into an HTTP GET Request
CHUNK_DOWNLOAD_RETRIES = 10
CHUNK_DOWNLOAD_TIMEOUT_S = 10


class ChunkSetStatus(Enum):
    """
    Chunk set status.

    Prevents the accidental access of chunk sets that have not yet received all of their data.
    """

    FILLING = 0
    EMPTYING = 1


def upload_chunks(api_address: str, chunks: List[Tuple[int, bytes]]) -> str:
    """
    Upload data chunks.

    Upload a series of data chunks through the chunked transfer mechanism.
    This method is for use on the client, not the server.

    Parameters
    ----------
    api_address : str
        The full URL of the API, including port but not including endpoint, e.g. "http://127.0.0.1:9999"
    chunks : List[Tuple[int, bytes]]
        A list of tuples containing the ordinal number of the chunk and each chunk

    Returns
    -------
    str
        The chunk set uuid of the uploaded chunks
    """
    # post the first chunk to start the upload
    response = requests.post(api_address + f"/chunk/upload/{chunks[0][0]}-of-{len(chunks)}", data=chunks[0][1])
    response.raise_for_status()
    log.debug("response from very first /chunk/upload was '%s'", response.json())
    chunk_set_uuid = response.json()["chunk_set_uuid"]

    # post the rest of the chunks in a random order
    for i, chunk in chunks[1:]:
        response = requests.post(api_address + f"/chunk/upload/{chunk_set_uuid}/{i}-of-{len(chunks)}", data=chunk)
        response.raise_for_status()
        log.debug("response from subsequent /chunk/upload/uuid call was '%s'", response.text)

    return chunk_set_uuid


def download_chunks(api_address: str, chunk_set_uuid: str, num_chunks: int) -> bytes:
    """
    Download data chunks.

    Download a series of data chunks sequentially through the chunked transfer mechanism.

    Parameters
    ----------
    api_address : str
        The full URL of the API, including port but not including endpoint, e.g. "http://127.0.0.1:9999"
    chunk_set_uuid : str
        The uuid of the chunk set to download.
    num_chunks : int
        The number of chunks to download.

    Returns
    -------
    bytes
        The concatenated data from all the chunks.

    Raises
    ------
    RuntimeError
        If the download fails after a number of retries.
    """
    chunks = []
    for chunk_ordinal in range(1, num_chunks + 1):
        retries = 0
        success = False
        while retries < CHUNK_DOWNLOAD_RETRIES and not success:
            try:
                this_chunk_response = requests.get(
                    f"{api_address}/chunk/download/{chunk_set_uuid}/{chunk_ordinal}", timeout=CHUNK_DOWNLOAD_TIMEOUT_S
                )
                this_chunk_response.raise_for_status()
                ordinal = struct.unpack("i", this_chunk_response.content[:4])[0]
                chunk = this_chunk_response.content[4:]
                assert ordinal == chunk_ordinal, f"Chunk ordinal mismatch: expected {chunk_ordinal}, got {ordinal}"
                chunks.append(chunk)
                success = True
            except (requests.Timeout, requests.ConnectionError) as e:
                retries += 1
                log.warning("Request failed with %s. Retrying (%s/%s)...", e, retries, CHUNK_DOWNLOAD_RETRIES)
                time.sleep(1)  # Small backoff before retrying
        if not success:
            error_message = (
                f"Failed to download from chunk set {chunk_set_uuid} after {CHUNK_DOWNLOAD_TIMEOUT_S} attempts."
            )
            log.error(error_message)
            raise RuntimeError(error_message)

    # Let the server know that we are done with this data and it can be deleted.
    requests.delete(f"{api_address}/chunk/delete/{chunk_set_uuid}")

    return b"".join(chunks)


def split_into_chunks(data: bytes, chunk_size: int = MAX_CHUNK_SIZE_BYTES) -> List[Tuple[int, bytes]]:
    """
    Split data into compressed chunks for transport.

    Parameters
    ----------
    data : bytes
        The data to be split into chunks.
    chunk_size: int
        The maximum number of bytes allowed in each chunk.

    Returns
    -------
    List[Tuple[int, bytes]]
        A list of tuples containing the ordinal number of the chunk and each chunk
    """
    return list(enumerate((data[i : i + chunk_size] for i in range(0, len(data), chunk_size)), start=1))


def deserialise_dataframe(data: bytes) -> Union[DataFrame, GeoDataFrame]:
    """
    Convert a binary serialised feather table to pandas dataframe.

    Parameters
    ----------
    data : bytes
        Binary serialised feather table.

    Returns
    -------
    Union[DataFrame, GeoDataFrame]
        Input table converted to a pandas dataframe.

    Raises
    ------
    ValueError
        When bytes can't be interpreted as meaningful dataframe.
    """
    try:
        buffer = BytesIO(data)
        df = read_geofeather(buffer)  # type: ignore
    except ValueError as e:
        # First attempt to deserialise as a geodataframe. If geo meta is missing, we expect a clear ValueError
        # and we then load as a plain dataframe instead.
        if "Missing geo meta" in e.args[0] or "'geo' metadata" in e.args[0]:
            try:
                df = read_feather(BytesIO(data))
            except ValueError as e:
                raise ValueError("Couldn't deserialise table format") from e
        else:
            raise ValueError("Couldn't deserialise table format") from e
    return df  # type: ignore


def serialise_dataframe(df: Union[DataFrame, GeoDataFrame]) -> bytes:
    """
    Serialise a dataframe using the feather table format.

    Parameters
    ----------
    df: Union[DataFrame, GeoDataFrame]
        Input dataframe

    Returns
    -------
    bytes
        Serialised feather table.
    """
    feather_buffer = BytesIO()
    # Browser implementations of feather do not support compressed feather formats.
    df.to_feather(feather_buffer, compression="uncompressed")
    feather_buffer.seek(0)
    return feather_buffer.getvalue()



"""Unit tests for the Specklia Client."""

import json
import struct
from datetime import datetime
from typing import Dict, List
from unittest.mock import patch

import geopandas as gpd
import pandas as pd
import pytest
import responses
from shapely import Polygon, to_geojson
from shapely.geometry import mapping

from specklia import Specklia, chunked_transfer

_QUERY_DATASET_DICT = {
    "dataset_id": "sheffield",
    "epsg4326_polygon": Polygon(((0, 0), (0, 1), (1, 1), (0, 0))),
    "min_timestamp": datetime(2000, 1, 1),
    "max_timestamp": datetime(2000, 1, 2),
    "columns_to_return": ["croissant"],
    "additional_filters": [
        {"column": "cheese", "operator": "<", "threshold": 6.57},
        {"column": "wine", "operator": ">=", "threshold": -23},
    ],
}


@pytest.fixture
def example_geodataframe() -> gpd.GeoDataFrame:
    return gpd.GeoDataFrame(
        {"geometry": gpd.points_from_xy([1, 2, 3, 4, 5], [0, 1, 2, 3, 4]), "timestamp": [2, 3, 4, 5, 6]},
        crs="EPSG:4326",
    )


@pytest.fixture
def example_datasets_dataframe() -> pd.DataFrame:
    return pd.DataFrame(
        [
            {
                "columns": {
                    "description": "hobbit height in cm",
                    "max_value": "150",
                    "min_value": "0",
                    "name": "height",
                    "type": "int",
                    "unit": "centimetres",
                },
                "created_timestamp": "Sat, 1 Jan 2000 15:44:24",
                "dataset_id": "sauron",
                "dataset_name": "hobbit_height",
                "description": "The height of some hobbits",
                "epsg4326_coverage": mapping(Polygon(((0, 1), (1, 1), (1, 0)))),
                "last_modified_timestamp": "Sun, 2 Jan 2000 15:44:24",
                "last_queried_timestamp": "Sun, 2 Jan 2000 12:14:44",
                "max_timestamp": "Wed, 1 Dec 1999 12:10:54",
                "min_timestamp": "Mon, 1 Nov 1999 00:41:25",
                "owning_group_id": "pippin",
                "owning_group_name": "merry",
                "size_rows": 4,
                "size_uncompressed_bytes": 726,
            }
        ]
    )


@pytest.fixture
def example_usage_report() -> List[Dict]:
    return [
        {
            "total_billable_bytes_processed": 10,
            "total_increase_in_bytes_stored": 20,
            "user_id": "example_user",
            "year": 2023,
            "month": 11,
        }
    ]


@pytest.fixture
def test_client(token: str) -> Specklia:
    with patch.object(Specklia, "_fetch_user_id"):
        return Specklia(auth_token=token, url="https://localhost")


def test_create_client(test_client: Specklia) -> None:
    assert test_client is not None


@responses.activate(assert_all_requests_are_fired=True)
def test_user_id(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(responses.POST, "https://localhost/users", json="fake_user_id", match=[token_matcher])
    test_client._fetch_user_id()
    assert test_client.user_id == "fake_user_id"


@responses.activate(assert_all_requests_are_fired=True)
def test_list_users(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.GET,
        "https://localhost/users",
        json=[{"name": "fred", "email": "fred@fred.fred"}],
        match=[token_matcher],
    )
    test_client.list_users(group_id="hazbin")


@responses.activate(assert_all_requests_are_fired=True)
def test_add_points_to_dataset(
    test_client: Specklia, example_geodataframe: gpd.GeoDataFrame, token_matcher: responses.matchers.header_matcher
) -> None:
    responses.add(
        responses.POST,
        "https://localhost/ingest",
        json={"chunk_set_uuid": "brian"},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher(
                {
                    "dataset_id": "dummy_dataset",
                    "new_points": [{"source": {"reference": "cheese"}, "chunk_set_uuid": "brian", "num_chunks": 1}],
                    "duplicate_source_behaviour": "error",
                },
            ),
        ],
    )
    responses.add(
        responses.POST,
        "https://localhost/chunk/upload/1-of-1",
        json={"chunk_set_uuid": "brian"},
    )
    test_client.add_points_to_dataset(
        dataset_id="dummy_dataset", new_points=[{"source": {"reference": "cheese"}, "gdf": example_geodataframe}]
    )


@responses.activate(assert_all_requests_are_fired=True)
def test_query_dataset(
    test_client: Specklia, example_geodataframe: gpd.GeoDataFrame, token_matcher: responses.matchers.header_matcher
) -> None:
    responses.add(
        responses.POST,
        "https://localhost/query",
        json={
            "chunk_set_uuid": "brian",
            "num_chunks": 1,
            "sources": [
                {
                    "geospatial_coverage": mapping(Polygon()),
                    "min_time": datetime.now().isoformat(),
                    "max_time": datetime.now().isoformat(),
                }
            ],
        },
        match=[
            token_matcher,
            responses.matchers.json_params_matcher(
                {
                    "dataset_id": "dummy_dataset",
                    "epsg4326_search_area": json.loads(to_geojson(Polygon(((0, 0), (0, 1), (1, 1), (0, 0))))),
                    "min_timestamp": 1588719600,
                    "max_timestamp": 1589065200,
                    "columns_to_return": ["lat", "lon"],
                    "additional_filters": [
                        {"column": "cheese", "operator": "<", "threshold": 6.57},
                        {"column": "wine", "operator": ">=", "threshold": -23},
                    ],
                    "source_information_only": False,
                }
            ),
        ],
    )

    responses.add(
        responses.GET,
        "https://localhost/chunk/download/brian/1",
        body=struct.pack("i", 1) + chunked_transfer.serialise_dataframe(example_geodataframe),
    )
    responses.add(responses.DELETE, "https://localhost/chunk/delete/brian")

    response = test_client.query_dataset(
        dataset_id="dummy_dataset",
        epsg4326_polygon=Polygon(((0, 0), (0, 1), (1, 1), (0, 0))),
        min_datetime=datetime(2020, 5, 6),
        max_datetime=datetime(2020, 5, 10),
        columns_to_return=["lat", "lon"],
        additional_filters=[
            {"column": "cheese", "operator": "<", "threshold": 6.57},
            {"column": "wine", "operator": ">=", "threshold": -23},
        ],
    )

    pd.testing.assert_frame_equal(response[0], example_geodataframe)


@pytest.mark.parametrize(
    ("invalid_json", "expected_exception", "expected_match"),
    # invalid espg4326_search_area type
    [
        (dict(_QUERY_DATASET_DICT, epsg4326_polygon="my back garden"), TypeError, "provide only Geometry objects"),
        # invalid min_datetime type
        (
            dict(_QUERY_DATASET_DICT, min_timestamp="a long time ago"),
            AttributeError,
            "object has no attribute 'timestamp'",
        ),
        # invalid max_datetime type
        (
            dict(_QUERY_DATASET_DICT, max_timestamp="the year 3000"),
            AttributeError,
            "object has no attribute 'timestamp'",
        ),
    ],
)
def test_query_dataset_invalid_request(
    test_client: Specklia, invalid_json: dict, expected_exception: type[Exception], expected_match: str
) -> None:
    with pytest.raises(expected_exception, match=expected_match):
        test_client.query_dataset(
            dataset_id=invalid_json["dataset_id"],
            epsg4326_polygon=invalid_json["epsg4326_polygon"],
            min_datetime=invalid_json["min_timestamp"],
            max_datetime=invalid_json["max_timestamp"],
            columns_to_return=invalid_json["columns_to_return"],
            additional_filters=invalid_json["additional_filters"],
        )


@responses.activate(assert_all_requests_are_fired=True)
def test_list_all_groups(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(responses.GET, "https://localhost/groups", json=["ducks"], match=[token_matcher])
    test_client.internal_admin.list_all_groups()


@responses.activate(assert_all_requests_are_fired=True)
def test_create_group(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.POST,
        "https://localhost/groups",
        json={},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher({"group_name": "ducks"}),
        ],
    )
    test_client.create_group("ducks")


@responses.activate(assert_all_requests_are_fired=True)
def test_update_group_name(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.PUT,
        "https://localhost/groups",
        json={},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher({"group_id": "ducks", "new_group_name": "pigeons"}),
        ],
    )
    test_client.update_group_name(group_id="ducks", new_group_name="pigeons")


@responses.activate(assert_all_requests_are_fired=True)
def test_delete_group(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.DELETE,
        "https://localhost/groups",
        json={},
        match=[
            token_matcher,
            responses.matchers.query_param_matcher({"group_id": "ducks"}),
        ],
    )
    test_client.delete_group(group_id="ducks")


@responses.activate(assert_all_requests_are_fired=True)
def test_list_groups(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(responses.GET, "https://localhost/groupmembership", json=["ducks"], match=[token_matcher])
    test_client.list_groups()


@responses.activate(assert_all_requests_are_fired=True)
def test_add_user_to_group(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.POST,
        "https://localhost/groupmembership",
        json={},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher({"group_id": "ducks", "user_to_add_id": "donald"}),
        ],
    )
    test_client.add_user_to_group(group_id="ducks", user_to_add_id="donald")


@responses.activate(assert_all_requests_are_fired=True)
def test_update_user_privileges(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.PUT,
        "https://localhost/groupmembership",
        json={},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher(
                {"group_id": "ducks", "user_to_update_id": "donald", "new_privileges": "ADMIN"}
            ),
        ],
    )
    test_client.update_user_privileges(group_id="ducks", user_to_update_id="donald", new_privileges="ADMIN")


@responses.activate(assert_all_requests_are_fired=True)
def test_delete_user_from_group(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.DELETE,
        "https://localhost/groupmembership",
        json={},
        match=[
            token_matcher,
            responses.matchers.query_param_matcher({"group_id": "ducks", "user_to_delete_id": "donald"}),
        ],
    )
    test_client.delete_user_from_group(group_id="ducks", user_to_delete_id="donald")


@responses.activate(assert_all_requests_are_fired=True)
def test_generate_api_key(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    user_dict = {"user_id": "user_1233", "token": "specklia_token_123"}
    responses.add(
        responses.PUT,
        "https://localhost/generate_user_api_key/user_1233",
        json=user_dict,
        match=[token_matcher],
    )
    response = test_client.internal_admin.generate_user_api_key(user_id="user_1233")
    assert response == user_dict


@responses.activate(assert_all_requests_are_fired=True)
def test_list_datasets(
    test_client: Specklia, token_matcher: responses.matchers.header_matcher, example_datasets_dataframe: pd.DataFrame
) -> None:
    responses.add(
        responses.GET,
        "https://localhost/metadata",
        json=example_datasets_dataframe.to_dict(orient="records"),
        match=[token_matcher],
    )
    datasets = test_client.list_datasets()
    assert type(datasets["epsg4326_coverage"][0]) is Polygon
    for column in datasets.columns:
        if "timestamp" in column:
            assert type(datasets[column][0]) is pd.Timestamp


@responses.activate(assert_all_requests_are_fired=True)
def test_create_dataset(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.POST,
        "https://localhost/metadata",
        json={},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher(
                {
                    "dataset_name": "am",
                    "description": "wibble",
                    "columns": [
                        {"name": "hobbits", "type": "halflings", "description": "concerning hobbits"},
                        {"name": "cats", "type": "pets", "description": "concerning cats"},
                    ],
                    "storage_technology": "OLAP",
                }
            ),
        ],
    )
    test_client.create_dataset(
        dataset_name="am",
        description="wibble",
        columns=[
            {"name": "hobbits", "type": "halflings", "description": "concerning hobbits"},
            {"name": "cats", "type": "pets", "description": "concerning cats"},
        ],
    )


@responses.activate(assert_all_requests_are_fired=True)
def test_update_dataset_ownership(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.PUT,
        "https://localhost/metadata",
        json={},
        match=[
            token_matcher,
            responses.matchers.json_params_matcher({"dataset_id": "bside", "new_owning_group_id": "arctic monkeys"}),
        ],
    )
    test_client.update_dataset_ownership(dataset_id="bside", new_owning_group_id="arctic monkeys")


@responses.activate(assert_all_requests_are_fired=True)
def test_delete_dataset(test_client: Specklia, token_matcher: responses.matchers.header_matcher) -> None:
    responses.add(
        responses.DELETE,
        "https://localhost/metadata",
        json={},
        match=[
            token_matcher,
            responses.matchers.query_param_matcher({"dataset_id": "bside"}),
        ],
    )
    test_client.delete_dataset(dataset_id="bside")


@responses.activate(assert_all_requests_are_fired=True)
def test_report_usage(
    test_client: Specklia, token_matcher: responses.matchers.header_matcher, example_usage_report: List[Dict]
) -> None:
    responses.add(
        responses.GET,
        "https://localhost/usage",
        json=example_usage_report,
        match=[
            token_matcher,
            responses.matchers.query_param_matcher({"group_id": "beatles"}),
        ],
    )
    group_id = "beatles"
    results = test_client.report_usage(group_id=group_id)
    assert len(results) == 1
    assert results[0]["user_id"] == "example_user"



"""Setup file for The specklia client."""

import os

from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

if os.path.exists("full_version.txt"):
    with open("full_version.txt", "r", encoding="utf-8") as fh:
        """
        Note that this file is generated by the CI chain based on the git tag
        (by ew_continuous_integration/define_new_version_number.py)
        It should not be present in the repository by default.
        """
        version_number = fh.read()
else:
    version_number = "v0.0.0"  # default value when under development

setup(
    name="specklia",
    version=version_number,
    description="Python client for Specklia, a geospatial point cloud database by Earthwave.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Earthwave Ltd",
    author_email="support@earthwave.co.uk",
    url="https://specklia.earthwave.co.uk/",
    python_requires=">=3.11",
    license="MIT",
    packages=find_packages(),
    # These generate the icons in the sidebar on PyPI
    project_urls={
        "Homepage": "https://specklia.earthwave.co.uk/",
        "Changelog": "https://specklia.earthwave.co.uk/generated_docs/change_log.html",
        "Documentation": "https://specklia.earthwave.co.uk/generated_docs/index.html",
        "Twitter": "https://twitter.com/earth__wave",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.11",
        "Topic :: Database :: Front-Ends",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
    ],
    # note requirements listed ininstall_requires should be the *minimum required*
    # in order to allow pip to resolve multiple installed packages properly.
    # requirements.txt should contain a specific known working version instead.
    install_requires=[
        "blosc",
        "geopandas",
        "pandas",
        "pyarrow",
        "rasterio",
        "requests",
        "shapely",
    ],
)
